"""
municipality
============
"""

__all__ = ("Municipality",)

import dataclasses
import json
import re
from threading import Thread

import requests

from ..utils import BakalibError, _setup_logger, data_dir


class Municipality:
    """
    Provides info about all schools that use the Bakaláři system.\n
        >>> m = Municipality()
        >>> for city in m.municipality().cities:
        >>>     print(city.name)
        >>>     for school in city.schools:
        >>>         print(school.name)
        >>>         print(school.domain)
    Methods:\n
            municipality(): Returns your local database as a NamedTuple
            build(): Builds the local database from 'https://sluzby.bakalari.cz/api/v1/municipality'.
                     Library comes prepackaged with a database json. Use only when needed.
    """

    db_file = data_dir.joinpath("municipality", "db.json")

    @dataclasses.dataclass(frozen=True)
    class Result:
        cities: list

        def __len__(self):
            return len(self.cities)

    @dataclasses.dataclass(frozen=True)
    class City:
        name: str
        school_count: str
        schools: list

        def __len__(self):
            return len(self.schools)

    @dataclasses.dataclass(frozen=True)
    class School:
        id: str
        name: str
        domain: str

    def __init__(self):
        self.logger = _setup_logger("municipality")

        self.thread = Thread(target=self._municipality)
        self.thread.start()
        self.logger.info("MUNICIPALITY THREAD STARTED")

    def municipality(self):
        if self.thread.is_alive():
            self.logger.info("MUNICIPALITY THREAD RUNNING")
            self.thread.join()
            self.logger.info("MUNICIPALITY THREAD FINISHED")
        return self._municipality()

    def _municipality(self):
        if not data_dir.is_dir():
            data_dir.mkdir()
        if self.db_file.is_file():
            db = json.loads(self.db_file.read_text(encoding="utf-8"), encoding="utf-8")
            result = self.Result(
                [
                    self.City(
                        name=city["name"],
                        school_count=city["school_count"],
                        schools=[
                            self.School(
                                id=school["id"],
                                name=school["name"],
                                domain=school["domain"],
                            )
                            for school in city["schools"]
                        ],
                    )
                    for city in db["cities"]
                ]
            )
            return result
        else:
            return self.build()

    def build(self):
        import lxml.etree as ET

        url = "https://sluzby.bakalari.cz/api/v1/municipality/"
        parser = ET.XMLParser(recover=True)

        try:
            result = self.Result(
                [
                    self.City(
                        municInfo.find("name").text,
                        municInfo.find("schoolCount").text,
                        [
                            self.School(
                                school.find("id").text,
                                school.find("name").text,
                                re.sub(
                                    "((/)?login.aspx(/)?)?",
                                    "",
                                    re.sub(
                                        "http(s)?://(www.)?",
                                        "",
                                        school.find("schoolUrl").text,
                                    ),
                                ).rstrip("/"),
                            )
                            for school in ET.fromstring(
                                requests.get(
                                    url
                                    + requests.utils.quote(municInfo.find("name").text),
                                    stream=True,
                                ).content,
                                parser=parser,
                            ).iter("schoolInfo")
                            if school.find("name").text
                        ],
                    )
                    for municInfo in ET.fromstring(
                        requests.get(url, stream=True).content, parser=parser
                    ).iter("municipalityInfo")
                    if municInfo.find("name").text
                ]
            )
        except Exception as e:
            self.logger.error(f"{type(e)}: {e}")
            raise BakalibError("Municipality failed to generate")

        self.db_file.write_text(
            json.dumps(dataclasses.asdict(result), indent=4, sort_keys=True),
            encoding="utf-8",
        )
        return result



"""
timetable
=========
"""

__all__ = ("Timetable",)

import datetime
import html
from concurrent.futures.thread import ThreadPoolExecutor
from dataclasses import dataclass

from ..core.client import Client
from ..utils import _setup_logger, cache, request
from ._generic import Generic


class Timetable(Generic):
    """
    Obtains information from the "rozvrh" module of Bakaláři.
    >>> timetable = Timetable(url, token)
    >>> timetable = Timetable(client) # <- You can also use a `Client` instance
    Methods:
        prev_week(prune: bool): Decrements self.date by 7 days and points to self.date_week.
        this_week(prune: bool): Points to date_week() with current date.
        next_week(prune: bool): Increments self.date by 7 days and points to self.date_week.
        date_week(date: datetime.date, prune: bool): Obtains timetable data about the week of the provided date.
        clear_cache(): Clears cache.
    """

    def __init__(
        self,
        client: Client = None,
        url: str = None,
        token: str = None,
        date: datetime.date = datetime.date.today(),
    ):
        super().__init__(client=client, url=url, token=token)
        self.date = date
        self.logger = _setup_logger(f"timetable_{client.username}")

        self.threadpool = ThreadPoolExecutor(max_workers=8)
        self.logger.info("TIMETABLE THREADPOOL CREATED")
        self.threadpool.submit(self._date_week, self.date)
        self.logger.info("TASK SUBMITTED")

    # ----------------------------------------------------

    def prev_week(self, prune: bool = True) -> "Timetable._date_week.Result":
        self.date = self.date - datetime.timedelta(7)
        return self.date_week(self.date, prune=prune)

    def this_week(self, prune: bool = True) -> "Timetable._date_week.Result":
        self.date = datetime.date.today()
        return self.date_week(self.date, prune=prune)

    def next_week(self, prune: bool = True) -> "Timetable._date_week.Result":
        self.date = self.date + datetime.timedelta(7)
        return self.date_week(self.date, prune=prune)

    # ----------------------------------------------------

    def date_week(
        self, date: datetime.date = None, prune: bool = True
    ) -> "Timetable._date_week.Result":
        """
        Obtains all timetable data about the week of the provided date.
        >>> this_week = timetable.date_week(datetime.date.today())
        >>> this_week = timetable.date_week(datetime.date.today(), prune=False) # <- Use this if you want to preserve empty lessons.
        >>> for header in this_week.headers:
        >>>     header.caption
        >>> for day in this_week.days:
        >>>     day.abbr
        >>>     for lesson in day.lessons:
        >>>         lesson.name
        >>>         lesson.teacher
        """
        self.date = date if date else self.date
        date_str = "{:04}{:02}{:02}".format(
            self.date.year, self.date.month, self.date.day
        )

        if not (self.url, self.token, "rozvrh", date_str) in cache:
            self.threadpool.shutdown(wait=True)
            self.logger.info("STOPPED, CREATING NEW THREADPOOL")
            self.threadpool = ThreadPoolExecutor(max_workers=8)
            self.logger.info("NEW THREADPOOL CREATED")

        self.threadpool.submit(self._date_week, self.date - datetime.timedelta(7))
        self.threadpool.submit(self._date_week, self.date + datetime.timedelta(7))
        self.logger.info("NEW TASKS SUBMITTED")

        return self._date_week(self.date, prune=prune)

    def _date_week(
        self, date: datetime.date, prune: bool = True
    ) -> "Timetable._date_week.Result":
        date_str = "{:04}{:02}{:02}".format(date.year, date.month, date.day)

        response = request(url=self.url, hx=self.token, pm="rozvrh", pmd=date_str)

        @dataclass(frozen=True)
        class Result:
            headers: list
            days: list
            cycle_name: str

            def __len__(self):
                return len(self.days)

        @dataclass(frozen=True)
        class Header:
            caption: str
            time_begin: str
            time_end: str

        @dataclass(frozen=True)
        class Day:
            abbr: str
            date: str
            lessons: list

            def __len__(self):
                return len(self.lessons)

        @dataclass(frozen=True)
        class Lesson:
            id_code: str
            type: str
            holiday: str
            abbr: str
            name: str
            teacher_abbr: str
            teacher: str
            room_abbr: str
            room: str
            absence_abbr: str
            absence: str
            theme: str
            group_abbr: str
            group: str
            cycle: str
            disengaged: str
            change_description: str
            notice: str
            caption: str
            time_begin: str
            time_end: str

        def holiday_check(obj: dict):
            zkratka = obj.get("zkratka")
            nazev = obj.get("nazev")
            if not zkratka:
                return nazev
            elif not nazev:
                return zkratka
            elif zkratka and nazev:
                return max([obj.get("zkratka"), obj.get("nazev")], key=len)
            else:
                return None

        headers = [
            Header(header["caption"], header["begintime"], header["endtime"])
            for header in response["rozvrh"]["hodiny"]["hod"]
        ]
        days = [
            Day(
                day["zkratka"],
                day["datum"],
                [
                    Lesson(
                        lesson.get("idcode"),
                        lesson.get("typ"),
                        holiday_check(lesson),
                        lesson.get("zkrpr"),
                        lesson.get("pr"),
                        lesson.get("zkruc"),
                        lesson.get("uc"),
                        lesson.get("zkrmist"),
                        lesson.get("mist"),
                        lesson.get("zkrabs"),
                        lesson.get("abs"),
                        html.unescape(lesson.get("tema"))
                        if lesson.get("tema")
                        else lesson.get("tema"),
                        lesson.get("zkrskup"),
                        lesson.get("skup"),
                        lesson.get("cycle"),
                        lesson.get("uvol"),
                        lesson.get("chng"),
                        lesson.get("notice"),
                        header.caption,
                        header.time_begin,
                        header.time_end,
                    )
                    for header, lesson in zip(headers, day["hodiny"]["hod"])
                ],
            )
            for day in response["rozvrh"]["dny"]["den"]
        ]

        if prune:
            lengths = []
            placeholder_lesson = None
            for day in days:
                for lesson in day.lessons:
                    if (
                        (lesson.type == "X" or lesson.type == "A")
                        and not lesson.holiday
                        and not lesson.change_description
                    ):
                        day.lessons.pop(0)
                    else:
                        break
                for lesson in reversed(day.lessons):
                    if (
                        (lesson.type == "X" or lesson.type == "A")
                        and not lesson.holiday
                        and not lesson.change_description
                    ):
                        day.lessons.pop()
                        placeholder_lesson = lesson
                    else:
                        break
                lengths.append(len(day))

            longest_day = None
            for day in sorted(days, key=len, reverse=True):
                if day.lessons[-1].type != "A" and day.lessons[-1].type != "X":
                    longest_day = day
                    break
            else:
                longest_day = max(days, key=len)

            headers = [
                Header(lesson.caption, lesson.time_begin, lesson.time_end)
                for lesson in (
                    longest_day.lessons if longest_day else max(days, key=len)
                )
            ]

            for day in days:
                while len(day.lessons) < max(lengths):
                    day.lessons.append(placeholder_lesson)

        return Result(headers, days, response["rozvrh"]["nazevcyklu"])

    def clear_cache(self) -> None:
        """
        Clears all entries related to the "rozvrh" module from global cache.
        >>> timetable.clear_cache()
        """
        for entry in cache:
            if "rozvrh" in entry:
                cache.pop(entry)



"""
Bakalib - library for accessing the Bakaláři school system easily!
core:
    client - authentication and info about the user
extra:
    municipality - list of cities and schools
modules:
    generic - boilerplate module
    grades
    timetable

"""

from ._version import version

__all__ = ("version",)



"""
util
====
"""

__all__ = (
    "BakalibError",
    "cache",
    "data_dir",
    "logger",
    "request",
)

import inspect as _inspect
import logging as _logging
import pathlib as _pathlib
import sys as _sys
import datetime as _datetime

import cachetools as _cachetools
import requests as _requests
import urllib3 as _urllib3
import xmltodict as _xmltodict

_urllib3.disable_warnings(_urllib3.exceptions.InsecureRequestWarning)

data_dir = _pathlib.Path(__file__).parent.joinpath("data")
_log_dir = _pathlib.Path(_pathlib.Path.home() / ".bakalib")
_log_file = _pathlib.Path(
    _log_dir / (_datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S") + ".log")
)

if not _log_dir.is_dir():
    _log_dir.mkdir()
else:
    _logs = [f for f in _log_dir.iterdir()]
    _logs.sort()
    while len(_logs) > 5:
        _logs[0].unlink()
        _logs.pop(0)

_log_file.touch()


def _setup_logger(name):
    formatter = _logging.Formatter(
        fmt="%(asctime)s — %(name)s — %(levelname)s — %(message)s — %(funcName)s:%(lineno)d"
    )
    handler = _logging.FileHandler(_log_file, mode="a")
    handler.setFormatter(formatter)
    logger = _logging.getLogger(name)
    logger.setLevel(_logging.DEBUG)
    logger.addHandler(handler)
    return logger


logger = _setup_logger("default")

cache = _cachetools.TTLCache(32, 300)


@_cachetools.cached(cache)
def request(url: str, **kwargs) -> dict:
    """
    Make a GET request to school URL.\n
    Module names are available at `https://github.com/bakalari-api/bakalari-api/tree/master/moduly`.
    >>> # Valid types of requests
    >>> request("https://example.com/login.aspx", gethx="Username123")
    >>> request("https://example.com/login.aspx", hx="token1234=", pm="module_name", pmd="20191219") # pmd is optional
    """

    r = _requests.get(url=url, params=kwargs, verify=False)
    response = _xmltodict.parse(r.content)
    results = response.get("results")

    try:
        res = results.get("res")
        result = results.get("result")
    except AttributeError as e:
        log_args = dict(kwargs)
        log_args.pop(k="hx", d=None)
        logger.error(log_args)

    comp = res if res else result

    if not comp == "01":
        raise BakalibError("Received response is invalid")

    return results


class BakalibError(Exception):
    pass



"""
client module

contains:

    Client: class

"""

__all__ = ("Client",)

import base64
import datetime
import hashlib
from dataclasses import dataclass
from threading import Thread

from ..utils import BakalibError, _setup_logger, request


def _is_logged_in_decorator(invert: bool = False):
    def decorator(func):
        def wrapper(*args, **kwargs):
            if args[0].logged_in ^ invert:
                return func(*args, **kwargs)
            else:
                if invert:
                    raise BakalibError("Client is already logged in.")
                raise BakalibError("Client is not logged in.")

        return wrapper

    return decorator


class Client:
    """
    Creates an instance with access to basic information of the user

    Parameters
    ----------
    username : str
        Client username recognizable by school server.
    domain : str
        School server domain (without 'http(s)', 'login.aspx').

    Attributes
    ----------
    username
    domain
    url : str
        School server URL generated from domain.
    perm_token : str
        Permanent token used for 'Remember password' function.
    token : str
        Token used for accessing the server.


    Methods
    -------
    login(password/perm_token: str = None)
        Logs the user in.
    info()
        Obtains user information (school name, year, ...)
    """

    def __init__(self, username: str, domain: str):
        super().__init__()
        self.username = username
        self.domain = domain
        self.url = f"https://{self.domain}/login.aspx"

        self.perm_token = None
        self.token = None
        self.thread = Thread()
        self.logger = _setup_logger(f"client_{self.username}")

        self.logged_in = False

        self.logger.info(f"CLIENT CREATED")

    def __str__(self):
        return f"Client(username={self.username}, domain={self.domain})"

    @_is_logged_in_decorator(invert=True)
    def login(
        self, password: str = None, perm_token: str = None, check_valid: bool = True
    ):
        if perm_token:
            self.perm_token = perm_token
            token = self._token(self.perm_token)
        elif password:
            self.perm_token = self._permanent_token(self.username, password)
            token = self._token(self.perm_token)
        else:
            raise BakalibError("Incorrect arguments")

        if check_valid:
            if self._is_token_valid(token):
                self.token = token
            else:
                raise BakalibError("Token is invalid: Invalid password/perm_token")
        else:
            self.token = token

        self.thread = Thread(
            target=request, args=(self.url,), kwargs={"hx": self.token, "pm": "login"}
        )
        self.thread.start()
        self.logger.info(f"INFO THREAD STARTED")

        self.logged_in = True
        self.logger.info(f"LOGGED IN SUCCESSFULLY")

    def _permanent_token(self, user: str, password: str) -> str:
        """
        Generates a permanent access token with securely hashed password.
        """
        self.logger.debug("GENERATING PERM_TOKEN")
        try:
            res = request(url=self.url, gethx=user)
        except BakalibError as e:
            print(e)
            raise BakalibError("Invalid username")

        salt = res.get("salt")
        ikod = res.get("ikod")
        typ = res.get("typ")
        salted_password = (salt + ikod + typ + password).encode("utf-8")
        hashed_password = base64.b64encode(hashlib.sha512(salted_password).digest())
        perm_token = (
            "*login*" + user + "*pwd*" + hashed_password.decode("utf8") + "*sgn*ANDR"
        )
        self.logger.debug("PERM_TOKEN GENERATED")
        return perm_token

    def _token(self, perm_token: str) -> str:
        """
        Generates an access token using current time.
        """
        self.logger.debug("GENERATING ACCESS TOKEN")
        today = datetime.date.today()
        datecode = "{:04}{:02}{:02}".format(today.year, today.month, today.day)
        token_hash = hashlib.sha512((perm_token + datecode).encode("utf-8")).digest()
        token = base64.urlsafe_b64encode(token_hash).decode("utf-8")
        self.logger.debug("ACCESS TOKEN GENERATED")
        return token

    def _is_token_valid(self, token: str) -> bool:
        """
        Checks for token validity.
        """
        try:
            request(url=self.url, hx=token, pm="login")
            self.logger.info(f"TOKEN IS VALID")
            return True
        except BakalibError:
            self.logger.warning(f"TOKEN IS INVALID")
            return False

    @_is_logged_in_decorator()
    def info(self) -> "Client.info.Result":
        """
        Obtains basic information about the user into a NamedTuple.
        >>> user.info().name
        >>> user.info().class_ # <-- due to class being a reserved keyword.
        >>> user.info().school
        """
        if self.thread.is_alive():
            self.logger.info(f"INFO THREAD RUNNING")
            self.thread.join()
            self.logger.info(f"INFO THREAD FINISHED")

        @dataclass(frozen=True)
        class Result:
            version: str
            name: str
            type_abbr: str
            type: str
            school: str
            school_type: str
            class_: str
            year: str
            modules: str
            newmarkdays: str

        response = request(url=self.url, hx=self.token, pm="login")
        result = Result(
            *[
                response.get(element).get("newmarkdays")
                if element == "params"
                else response.get(element)
                for element in response
                if not element == "result"
            ]
        )
        return result
